import contextlib
import random
import typing

from werkzeug.http import dump_cookie, parse_cookie

from apistar import http
from apistar.interfaces import SessionStore

local_memory_sessions = {}  # type: typing.Dict[str, typing.Dict[str, typing.Any]]


class LocalMemorySessionStore(SessionStore):
    cookie_name = 'session_id'

    def new(self) -> http.Session:
        session_id = self._generate_key()
        return http.Session(session_id=session_id)

    def load(self, session_id: str) -> http.Session:
        try:
            data = local_memory_sessions[session_id]
        except KeyError:
            return self.new()
        return http.Session(session_id=session_id, data=data)

    def save(self, session: http.Session) -> typing.Dict[str, str]:
        headers = {}
        if session.is_new:
            cookie = dump_cookie(self.cookie_name, session.session_id)
            headers['set-cookie'] = cookie
        if session.is_new or session.is_modified:
            local_memory_sessions[session.session_id] = session.data
        return headers

    def _generate_key(self) -> str:
        length = 30
        allowed_chars = 'abcdefghijklmnopqrstuvwxyz0123456789'
        urandom = random.SystemRandom()
        return ''.join(urandom.choice(allowed_chars) for i in range(length))


@contextlib.contextmanager
def get_session(cookie: http.Header,
                store: SessionStore,
                response_headers: http.ResponseHeaders) -> typing.Generator[http.Session, None, None]:
    if cookie:
        cookies = parse_cookie(cookie)
        session_id = cookies.get('session_id')
    else:
        session_id = None

    if session_id is not None:
        session = store.load(session_id)
    else:
        session = store.new()

    try:
        yield session
    finally:
        session_headers = store.save(session)
        response_headers.update(session_headers)



from apistar import Settings, exceptions, http
from apistar.interfaces import Injector
from apistar.renderers import DEFAULT_RENDERERS, negotiate_renderer
from apistar.types import Handler, ReturnValue


def check_permissions(handler: Handler,
                      injector: Injector,
                      settings: Settings) -> None:
    """
    Ensure that any configured permissions are met.

    Used in the default `BEFORE_REQUEST` configuration.
    """
    default_permissions = settings.get('PERMISSIONS', None)
    permissions = getattr(handler, 'permissions', default_permissions)
    if permissions is None:
        return

    for permission in permissions:
        if not injector.run(permission.has_permission):
            raise exceptions.Forbidden()


async def check_permissions_async(handler: Handler,
                                  injector: Injector,
                                  settings: Settings):
    """
    An async variant of 'check_permissions'.
    """
    default_permissions = settings.get('PERMISSIONS', None)
    permissions = getattr(handler, 'permissions', default_permissions)
    if permissions is None:
        return

    for permission in permissions:
        if not await injector.run_async(permission.has_permission):
            raise exceptions.Forbidden()


def render_response(handler: Handler,
                    injector: Injector,
                    settings: Settings,
                    accept: http.Header,
                    ret: ReturnValue) -> http.Response:
    """
    Render a response, using content negotiation to determine an
    appropriate renderer.

    Used in the default `BEFORE_REQUEST` configuration.
    """
    if isinstance(ret, http.Response):
        data, status, headers, content_type = ret
        if content_type is not None or (status >= 300 and status <= 399):
            return ret
    else:
        data, status, headers, content_type = ret, 200, {}, None

    if data is None or data == b'':
        content = b''
        content_type = None
    else:
        default_renderers = settings.get('RENDERERS', DEFAULT_RENDERERS)
        renderers = getattr(handler, 'renderers', default_renderers)
        renderer = negotiate_renderer(accept, renderers)
        if renderer is None:
            raise exceptions.NotAcceptable()
        content = injector.run(renderer.render, {'response_data': data})
        if isinstance(content, http.Response):
            return content
        content_type = renderer.get_content_type()

    if not content and status == 200:
        status = 204
        content_type = None

    return http.Response(content, status, headers, content_type)



"""
              _    ____ ___   ____  _
 __/\__      / \  |  _ \_ _| / ___|| |_ __ _ _ __    __/\__
 \    /     / _ \ | |_) | |  \___ \| __/ _` | '__|   \    /
 /_  _\    / ___ \|  __/| |   ___) | || (_| | |      /_  _\
   \/     /_/   \_\_|  |___| |____/ \__\__,_|_|        \/
"""
import importlib.util
import os
import sys

from apistar import exceptions
from apistar.core import Command, Component, Include, Route, annotate
from apistar.frameworks.cli import CliApp
from apistar.http import Response
from apistar.interfaces import App
from apistar.test import TestClient
from apistar.types import Settings

__version__ = '0.3.9'
__all__ = [
    'annotate', 'exceptions', 'Command', 'Component', 'Response', 'Route',
    'Include', 'Settings', 'TestClient'
]

_current_app = None


def main() -> None:
    if os.path.exists('app.py'):
        app = get_current_app()
    else:
        app = CliApp()
    app.main()


def get_current_app(use_cache=True) -> App:
    global _current_app

    if _current_app is not None and use_cache:
        return _current_app

    sys.path.insert(0, os.getcwd())
    spec = importlib.util.spec_from_file_location("app", "app.py")
    module = importlib.util.module_from_spec(spec)
    spec.loader.exec_module(module)
    app = getattr(module, 'app', None)
    if app is None:
        msg = "The 'app.py' module did not contain an 'app' variable."
        raise exceptions.ConfigurationError(msg)
    if not isinstance(app, App):
        msg = "The 'app' variable in 'app.py' is not an App instance."
        raise exceptions.ConfigurationError(msg)

    if use_cache:
        _current_app = app

    return app


def reverse_url(identifier: str, **values) -> str:
    app = get_current_app()
    return app.reverse_url(identifier, **values)


def render_template(template_name: str, **context) -> str:
    app = get_current_app()
    return app.render_template(template_name, **context)



from wsgiref.util import FileWrapper, request_uri

import werkzeug
from werkzeug.http import parse_options_header
from werkzeug.wsgi import get_input_stream

from apistar import Settings, exceptions, http, parsers
from apistar.authentication import Unauthenticated
from apistar.interfaces import Injector
from apistar.types import Handler, ParamName, WSGIEnviron


def get_method(environ: WSGIEnviron):
    return environ['REQUEST_METHOD'].upper()


def get_url(environ: WSGIEnviron):
    return http.URL(request_uri(environ))


def get_scheme(environ: WSGIEnviron):
    return environ['wsgi.url_scheme']


def get_host(environ: WSGIEnviron):
    return environ.get('HTTP_HOST') or environ['SERVER_NAME']


def get_port(environ: WSGIEnviron):
    if environ['wsgi.url_scheme'] == 'https':
        return int(environ.get('SERVER_PORT') or 443)
    return int(environ.get('SERVER_PORT') or 80)


def get_path(environ: WSGIEnviron):
    return environ['SCRIPT_NAME'] + environ['PATH_INFO']


def get_querystring(environ: WSGIEnviron):
    return environ.get('QUERY_STRING', '')


def get_queryparams(environ: WSGIEnviron) -> http.QueryParams:
    return werkzeug.urls.url_decode(
        environ.get('QUERY_STRING', ''),
        cls=http.QueryParams
    )


def get_queryparam(name: ParamName, queryparams: http.QueryParams):
    return queryparams.get(name)


def get_headers(environ: WSGIEnviron) -> http.Headers:
    header_items = []
    for key, value in environ.items():
        if key.startswith('HTTP_'):
            header = (key[5:].lower().replace('_', '-'), value)
            header_items.append(header)
        elif key in ('CONTENT_TYPE', 'CONTENT_LENGTH'):
            header = (key.lower().replace('_', '-'), value)
            header_items.append(header)
    return http.Headers(header_items)


def get_header(name: ParamName, headers: http.Headers):
    return headers.get(name.replace('_', '-'))


def get_body(environ: WSGIEnviron):
    return get_input_stream(environ).read()


def get_stream(environ: WSGIEnviron):
    return get_input_stream(environ)


def get_request_data(headers: http.Headers, injector: Injector, settings: Settings):
    content_type = headers.get('Content-Type')
    if not content_type:
        return None

    media_type, _ = parse_options_header(content_type)
    parser_mapping = {
        parser.media_type: parser
        for parser in settings.get('PARSERS', parsers.DEFAULT_PARSERS)
    }

    if media_type not in parser_mapping:
        raise exceptions.UnsupportedMediaType()

    parser = parser_mapping[media_type]
    return injector.run(parser.parse)


def get_auth(handler: Handler, injector: Injector, settings: Settings):
    default_authentication = settings.get('AUTHENTICATION', None)
    authentication = getattr(handler, 'authentication', default_authentication)
    if authentication is None:
        return Unauthenticated()

    for authenticator in authentication:
        auth = injector.run(authenticator.authenticate)
        if auth is not None:
            return auth
    return Unauthenticated()


def get_file_wrapper(environ: WSGIEnviron):
    return environ.get('wsgi.file_wrapper', FileWrapper)
